import unittest
import Tasmanian
import TasmanianSG # needed to test all rules
import numpy as np
import os, sys
from ctypes import cdll

from random import shuffle

import testConfigureData as tdata # needed to keep track of version tests and ways to load the library
import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Test make/update, check points, check library paths
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkPathsVersions(self):
        '''
        Check different ways to specify the library path and check the
        version number.
        '''
        grid = Tasmanian.TasmanianSparseGrid()

        sVersion = "{0:1d}.{1:1d}".format(grid.getVersionMajor(), grid.getVersionMinor())
        self.assertEqual(sVersion, Tasmanian.__version__, "version mismatch")
        sLicense = grid.getLicense()
        self.assertEqual(sLicense, Tasmanian.__license__, "license mismatch")

        iVM = int(sVersion.split('.')[0])
        iVm = int(sVersion.split('.')[1])
        self.assertEqual(iVM, grid.getVersionMajor(), "version major mismatch")
        self.assertEqual(iVm, grid.getVersionMinor(), "version minor mismatch")

    def checkMakeAgainstKnown(self):
        '''
        Make grids and check against pen-and-paper answers, check weights too.
        '''
        grid = Tasmanian.makeGlobalGrid(1, 0, 4, 'level', 'gauss-hermite', [], 2.0)
        aW = grid.getQuadratureWeights()
        self.assertTrue(np.abs(sum(aW) - 0.5 * np.sqrt(np.pi)) < 1.E-14, "Gauss-Hermite Alpha")

        grid.makeGlobalGrid(2, 0, 2, 'level', 'leja', [2, 1])
        aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0], [1.0, 0.0]])
        aP = grid.getPoints()
        np.testing.assert_equal(aA, aP, 'Anisotropy Global not equal', True)

        grid.makeGlobalGrid(2, 0, 4, 'ipcurved', 'leja', [20, 10, 0, 7])
        aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0], [0.0, np.sqrt(1.0/3.0)], [1.0, 0.0], [1.0, 1.0], [-1.0, 0.0]])
        aP = grid.getPoints()
        np.testing.assert_almost_equal(aA, aP, 12, 'Anisotropy Global not equal', True)

        grid.makeSequenceGrid(2, 1, 2, 'level', 'leja', [2, 1])
        aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0], [1.0, 0.0]])
        aP = grid.getPoints()
        np.testing.assert_equal(aA, aP, 'Anisotropy Sequence not equal', True)

        grid = Tasmanian.makeSequenceGrid(2, 1, 4, 'ipcurved', 'leja', [20, 10, 0, 7])
        aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0], [0.0, np.sqrt(1.0/3.0)], [1.0, 0.0], [1.0, 1.0], [-1.0, 0.0]])
        aP = grid.getPoints()
        np.testing.assert_almost_equal(aA, aP, 12, 'Anisotropy Sequence not equal', True)

        grid = Tasmanian.makeFourierGrid(2, 1, 2, 'level', [1, 2])
        aA = np.array([[0.0, 0.0], [0.0, 1.0/3.0], [0.0, 2.0/3.0], [1.0/3.0, 0.0], [2.0/3.0, 0.0], [1.0/9.0, 0.0], [2.0/9.0, 0.0], [4.0/9.0, 0.0], [5.0/9.0, 0.0], [7.0/9.0, 0.0], [8.0/9.0, 0.0]])
        aP = grid.getPoints()
        np.testing.assert_almost_equal(aA, aP, 12, 'Anisotropy Fourier not equal', True)

        # this is a very important test, checks the curved rule and covers the non-lower-set index selection code
        grid.makeGlobalGrid(2, 1, 1, 'ipcurved', 'rleja', [10, 10, -21, -21])
        aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5], [0.0, 0.25], [0.0, 0.75], [0.0, 0.125],
                       [1.0, 0.0], [1.0, 1.0], [1.0, 0.5], [1.0, 0.25], [1.0, 0.75], [1.0, 0.125],
                       [0.5, 0.0], [0.5, 1.0], [0.5, 0.5], [0.5, 0.25], [0.5, 0.75], [0.5, 0.125],
                       [0.25, 0.0], [0.25, 1.0], [0.25, 0.5], [0.25, 0.25], [0.25, 0.75],
                       [0.75, 0.0], [0.75, 1.0], [0.75, 0.5], [0.75, 0.25],
                       [0.125, 0.0], [0.125, 1.0], [0.125, 0.5]])
        aA = np.cos(np.pi * aA)
        aP = grid.getPoints()
        np.testing.assert_almost_equal(aA, aP, 12, 'Anisotropy heavily curved not equal', True) # 12 is the number of dec places

        # test weights
        aX = np.array([[0.0, -0.3], [-0.44, 0.7], [0.82, -0.01]])
        grid.makeGlobalGrid(2, 0, 4, 'level', 'chebyshev')
        aW = grid.getInterpolationWeightsBatch(aX)
        aX1 = aX[0,:].reshape([2,])
        aX2 = aX[1,:].reshape([2,])
        aX3 = aX[2,:].reshape([2,])
        aW1 = grid.getInterpolationWeights(aX1)
        aW2 = grid.getInterpolationWeights(aX2)
        aW3 = grid.getInterpolationWeights(aX3)
        aWW = np.row_stack([aW1, aW2, aW3])
        np.testing.assert_equal(aW, aWW, "Batch weights", True)

        grid.makeGlobalGrid(3, 0, 1, 'level', 'chebyshev')
        aTrans = np.array([[-1.0, -4.0], [2.0, 5.0], [-3.0, 5]])
        grid.setDomainTransform(aTrans)
        aA = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, -1.0], [0.0, 0.0, 1.0], [0.0, -1.0, 0.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0], [1.0, 0.0, 0.0]])
        aB = np.array([[-2.5, 3.5, 1.0], [-2.5, 3.5, -3.0], [-2.5, 3.5, 5.0], [-2.5,  2.0, 1.0], [-2.5, 5.0, 1.0], [-1.0, 3.5, 1.0], [-4.0, 3.5, 1.0]])
        np.testing.assert_equal(aB, grid.getPoints(), "Original equal", True)
        grid.clearDomainTransform()
        np.testing.assert_equal(aA, grid.getPoints(), "Tansformed equal", True)

        grid.makeGlobalGrid(3, 0, 1, 'level', 'fejer2')
        aA = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, -0.707106781186548], [0.0, 0.0, 0.707106781186548], [0.0, -0.707106781186548, 0.0], [0.0, 0.707106781186548, 0.0], [-0.707106781186548, 0.0, 0.0], [0.707106781186548, 0.0, 0.0]])
        np.testing.assert_almost_equal(aA, grid.getPoints(), 14, "Original equal", True)
        grid.setConformalTransformASIN(np.array([4, 6, 3]))
        aA = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, -0.60890205], [0.0, 0.0, 0.60890205], [0.0, -0.57892643, 0.0], [0.0, 0.57892643, 0.0], [-0.59587172, 0.0, 0.0], [0.59587172, 0.0, 0.0]])
        np.testing.assert_almost_equal(aA, grid.getPoints(), 7, "Transformed equal", True)
        grid.clearConformalTransform()
        aA = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, -0.707106781186548], [0.0, 0.0, 0.707106781186548], [0.0, -0.707106781186548, 0.0], [0.0, 0.707106781186548, 0.0], [-0.707106781186548, 0.0, 0.0], [0.707106781186548, 0.0, 0.0]])
        np.testing.assert_almost_equal(aA, grid.getPoints(), 14, "Original equal", True)

        # number of points
        grid = Tasmanian.makeLocalPolynomialGrid(2, 1, 2, 2, "localp")
        iNN = grid.getNumNeeded()
        iNL = grid.getNumLoaded()
        iNP = grid.getNumPoints()
        self.assertEqual(iNN, 13, "num needed")
        self.assertEqual(iNL, 0, "num loaded")
        self.assertEqual(iNP, iNN, "num points")
        ttc.loadExpN2(grid)
        iNN = grid.getNumNeeded()
        iNL = grid.getNumLoaded()
        iNP = grid.getNumPoints()
        self.assertEqual(iNN, 0, "num needed")
        self.assertEqual(iNL, 13, "num loaded")
        self.assertEqual(iNP, iNL, "num points")
        grid.setSurplusRefinement(0.10, 0, "classic")
        iNN = grid.getNumNeeded()
        iNL = grid.getNumLoaded()
        iNP = grid.getNumPoints()
        self.assertEqual(iNN, 8, "num needed")
        self.assertEqual(iNL, 13, "num loaded")
        self.assertEqual(iNP, iNL, "num points")
        grid.setSurplusRefinement(0.05, 0, "classic")
        iNN = grid.getNumNeeded()
        iNL = grid.getNumLoaded()
        iNP = grid.getNumPoints()
        self.assertEqual(iNN, 16, "num needed")
        self.assertEqual(iNL, 13, "num loaded")
        self.assertEqual(iNP, iNL, "num points")
        ttc.loadExpN2(grid)
        iNN = grid.getNumNeeded()
        iNL = grid.getNumLoaded()
        iNP = grid.getNumPoints()
        self.assertEqual(iNN,  0, "num needed")
        self.assertEqual(iNL, 29, "num loaded")
        self.assertEqual(iNP, iNL, "num points")

    def checkUpdate(self):
        '''
        Check the update and make working together.
        '''
        grid = Tasmanian.SparseGrid()
        for iI in range(2):
            if (iI == 0):
                sMake = "grid.makeGlobalGrid(2, 1, 2, 'level', 'leja', [2, 1])"
                sUpda1 = "grid.updateGlobalGrid(2, 'level', [2, 1])"
                sUpda2 = "grid.updateGlobalGrid(3, 'level', [2, 1])"
                sUpda3 = "grid.updateGlobalGrid(4, 'ipcurved', [20, 10, 0, 7])"
            else:
                sMake = "grid.makeSequenceGrid(2, 1, 2, 'level', 'leja', [2, 1])"
                sUpda1 = "grid.updateSequenceGrid(2, 'level', [2, 1])"
                sUpda2 = "grid.updateSequenceGrid(3, 'level', [2, 1])"
                sUpda3 = "grid.updateSequenceGrid(4, 'ipcurved', [20, 10, 0, 7])"
            exec(sMake)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 4, "num needed")
            self.assertEqual(iNL, 0, "num loaded")
            self.assertEqual(iNP, iNN, "num points")
            ttc.loadExpN2(grid)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 0, "num needed")
            self.assertEqual(iNL, 4, "num loaded")
            self.assertEqual(iNP, iNL, "num points")
            exec(sUpda1)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 0, "num needed")
            self.assertEqual(iNL, 4, "num loaded")
            self.assertEqual(iNP, iNL, "num points")
            exec(sUpda2)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 2, "num needed")
            self.assertEqual(iNL, 4, "num loaded")
            self.assertEqual(iNP, iNL, "num points")
            aA = np.array([[0.0, 0.0], [0.0, 1.0], [0.0, -1.0], [1.0, 0.0]])
            aP = grid.getPoints()
            np.testing.assert_equal(aA, aP, "Update Global not equal", True)
            aP = grid.getLoadedPoints()
            np.testing.assert_equal(aA, aP, "Update Global not equal", True)
            aA = np.array([[0.0, np.sqrt(1.0/3.0)], [1.0, 1.0]])
            aP = grid.getNeededPoints()
            np.testing.assert_equal(aA, aP, "Update Global not equal", True)
            ttc.loadExpN2(grid)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 0, "num needed")
            self.assertEqual(iNL, 6, "num loaded")
            self.assertEqual(iNP, iNL, "num points")
            exec(sUpda3)
            iNN = grid.getNumNeeded()
            iNL = grid.getNumLoaded()
            iNP = grid.getNumPoints()
            self.assertEqual(iNN, 1, "num needed")
            self.assertEqual(iNL, 6, "num loaded")
            self.assertEqual(iNP, iNL, "num points")
            aA = np.array([[-1.0, 0.0]])
            aP = grid.getNeededPoints()
            np.testing.assert_equal(aA, aP, "Update Global not equal", True)

        grid.makeFourierGrid(2, 1, 2, "level")
        ttc.loadExpN2(grid)
        grid.updateFourierGrid(3, "level")
        self.assertEqual(grid.getNumNeeded(), 60, "failed at updateFourierGrid()")

    def checkDefaults(self):
        '''
        Test that default values are set correctly for alpha/beta/order,
        empty returns, empty copy.
        '''
        grid = Tasmanian.SparseGrid()
        for sRule in TasmanianSG.lsTsgLocalRules:
            grid.makeLocalPolynomialGrid(3, 1, 3, 0, sRule)
            self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
            self.assertEqual(grid.getBeta(), 0.0, "failed beta")
            self.assertEqual(grid.getOrder(), 0, "failed order")
            grid.makeLocalPolynomialGrid(3, 1, 3, 1, sRule)
            self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
            self.assertEqual(grid.getBeta(), 0.0, "failed beta")
            self.assertEqual(grid.getOrder(), 1, "failed order")
            grid.makeLocalPolynomialGrid(3, 1, 3, 2, sRule)
            self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
            self.assertEqual(grid.getBeta(), 0.0, "failed beta")
            self.assertEqual(grid.getOrder(), 2, "failed order")

        grid = Tasmanian.makeWaveletGrid(3, 1, 2, 1)
        self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 0.0, "failed beta")
        self.assertEqual(grid.getOrder(), 1, "failed order")
        grid.makeWaveletGrid(3, 1, 2, 3)
        self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 0.0, "failed beta")
        self.assertEqual(grid.getOrder(), 3, "failed order")

        try:
            grid.copyGrid([])
            self.assertTrue(False, "failed to raise exception on copy grid")
        except Tasmanian.InputError as TsgError:
            TsgError.bShowOnExit = False

        # default alpha/beta and order
        grid.makeGlobalGrid(2, 0, 2, 'level', 'leja', [2, 1])
        self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 0.0, "failed beta")
        self.assertEqual(grid.getOrder(), -1, "failed order")
        grid.makeGlobalGrid(1, 0, 4, 'level', 'gauss-hermite', [], 2.0)
        self.assertEqual(grid.getAlpha(), 2.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 0.0, "failed beta")
        self.assertEqual(grid.getOrder(), -1, "failed order")
        grid.makeGlobalGrid(1, 0, 4, 'level', 'gauss-jacobi', [], 3.0, 2.0)
        self.assertEqual(grid.getAlpha(), 3.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 2.0, "failed beta")
        self.assertEqual(grid.getOrder(), -1, "failed order")

        grid.makeSequenceGrid(2, 1, 1, 'level', 'leja', [2, 1])
        self.assertEqual(grid.getAlpha(), 0.0, "failed alpha")
        self.assertEqual(grid.getBeta(), 0.0, "failed beta")
        self.assertEqual(grid.getOrder(), -1, "failed order")

        # test empty returns
        dummy_ans = np.empty([0, 0], np.float64)
        grid_dummy = Tasmanian.SparseGrid()
        aX = np.empty([0,], np.float64)
        np.testing.assert_equal(dummy_ans, grid_dummy.getPoints(), "Empty read", True)
        np.testing.assert_equal(dummy_ans, grid_dummy.getNeededPoints(), "Empty read", True)
        np.testing.assert_equal(dummy_ans, grid_dummy.getLoadedPoints(), "Empty read", True)
        dummy_ans = np.empty([0,], np.float64)
        np.testing.assert_equal(dummy_ans, grid_dummy.getQuadratureWeights(), "Empty read", True)
        np.testing.assert_equal(dummy_ans, grid_dummy.getInterpolationWeights(aX), "Empty read", True)
        dummy_ans = np.empty([0, 0], np.float64)
        aX = np.empty([0, 0], np.float64)
        np.testing.assert_equal(dummy_ans, grid_dummy.getInterpolationWeightsBatch(aX), "Empty read", True)
        aX = np.empty([1, 0], np.float64)
        np.testing.assert_equal(dummy_ans, grid_dummy.getInterpolationWeightsBatch(aX), "Empty read", True)
        grid.makeGlobalGrid(2, 1, 2, 'level', 'chebyshev')
        ttc.loadExpN2(grid)
        dummy_ans = np.empty([0, 1], np.float64)
        aX = np.empty([0, 2], np.float64)
        np.testing.assert_equal(dummy_ans, grid.evaluateBatch(aX), "Empty batch eval", True)

    def checkLevelLimits(self):
        '''
        Check setting level limits.
        '''
        grid = Tasmanian.SparseGrid()

        # Level Limits
        grid.makeGlobalGrid(2, 1, 2, 'level', 'clenshaw-curtis', liLevelLimits = [1, 4])
        aPoints = grid.getPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [-1.0, 1.0], lMustNotHave = [1.0/np.sqrt(2.0), -1.0/np.sqrt(2.0)])
        ttc.checkPoints(aPoints[:,1], lMustHave = [-1.0, 1.0, 1.0/np.sqrt(2.0), -1.0/np.sqrt(2.0)], lMustNotHave = [])
        ttc.loadExpN2(grid)
        grid.setAnisotropicRefinement('iptotal', 20, 0, [1, 4])
        self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [], lMustNotHave = [1.0/np.sqrt(2.0), -1.0/np.sqrt(2.0)])

        grid.makeSequenceGrid(3, 1, 3, 'level', 'leja', liLevelLimits = [3, 2, 1])
        aPoints = grid.getPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [0.0, -1.0, 1.0, 1.0/np.sqrt(3.0)], lMustNotHave = [])
        ttc.checkPoints(aPoints[:,1], lMustHave = [0.0, -1.0, 1.0], lMustNotHave = [1.0/np.sqrt(3.0)])
        ttc.checkPoints(aPoints[:,2], lMustHave = [0.0, 1.0], lMustNotHave = [-1.0, 1.0/np.sqrt(3.0)])
        ttc.loadExpN2(grid)
        grid.setAnisotropicRefinement('iptotal', 5, 0)
        self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [1.0/np.sqrt(3.0)])
        ttc.checkPoints(aPoints[:,2], lMustHave = [], lMustNotHave = [-1.0, 1.0/np.sqrt(3.0)])
        grid.clearRefinement()
        grid.setAnisotropicRefinement('iptotal', 10, 0, [3, 2, 2])
        self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [1.0/np.sqrt(3.0)])
        ttc.checkPoints(aPoints[:,2], lMustHave = [1.0], lMustNotHave = [1.0/np.sqrt(3.0)])
        grid.clearRefinement()
        grid.setSurplusRefinement(1.E-8, 0, '', [3, 2, 1])
        self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [1.0/np.sqrt(3.0)])
        ttc.checkPoints(aPoints[:,2], lMustHave = [], lMustNotHave = [-1.0, 1.0/np.sqrt(3.0)])

        # check that nodes from level 2 (+-0.5) and 3 (+-0.25, +-0.75) appear only in the proper dimension
        grid.makeLocalPolynomialGrid(3, 1, 3, 1, 'localp', [1, 2, 3])
        aPoints = grid.getPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [0.0, -1.0, 1.0], lMustNotHave = [0.5, -0.5, -0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,1], lMustHave = [0.0, -1.0, 1.0, 0.5, -0.5], lMustNotHave = [-0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,2], lMustHave = [0.0, -1.0, 1.0, 0.5, -0.5, -0.75, -0.25, 0.25, 0.75], lMustNotHave = [])
        ttc.loadExpN2(grid)
        grid.setSurplusRefinement(1.E-8, 0, 'classic')
        self.assertTrue(grid.getNumNeeded() > 0, 'did not refine')
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [], lMustNotHave = [0.5, -0.5, -0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [-0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,2], lMustHave = [], lMustNotHave = [])
        grid.clearRefinement()
        grid.setSurplusRefinement(1.E-8, 0, 'classic', [2, 2, 3])
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [0.5, -0.5], lMustNotHave = [-0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [-0.75, -0.25, 0.25, 0.75])

        grid.makeWaveletGrid(2, 1, 3, 1, [0, 2])
        aPoints = grid.getPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [0.0, -1.0, 1.0], lMustNotHave = [-0.5, 0.5])
        ttc.checkPoints(aPoints[:,1], lMustHave = [0.0, -1.0, 1.0, -0.5, 0.5, -0.75, -0.25, 0.25, 0.75], lMustNotHave = [-0.125, 0.125])
        ttc.loadExpN2(grid)
        grid.setSurplusRefinement(1.E-8, 0, 'classic') # grid is effectively full tensor, cannot refine within the level limits
        self.assertTrue((grid.getNumNeeded() == 0), 'added points even though it should not have');
        grid.setSurplusRefinement(1.E-8, 0, 'classic', [1, 2])
        aPoints = grid.getNeededPoints()
        ttc.checkPoints(aPoints[:,0], lMustHave = [0.5, -0.5], lMustNotHave = [-0.75, -0.25, 0.25, 0.75])
        ttc.checkPoints(aPoints[:,1], lMustHave = [], lMustNotHave = [-0.125, 0.125])

        # level limits I/O
        grid.makeLocalPolynomialGrid(3, 1, 3, 1, 'localp', [1, 2, 3])
        aLimits = grid.getLevelLimits()
        np.testing.assert_almost_equal(aLimits, np.array([1, 2, 3]), 14, "Could not read level limits", True)

        grid.clearLevelLimits()
        aLimits = grid.getLevelLimits()
        np.testing.assert_almost_equal(aLimits, np.array([-1, -1, -1]), 14, "Could not read level limits", True)

    def performMakeUpdateTests(self):
        self.checkPathsVersions()
        self.checkMakeAgainstKnown()
        self.checkUpdate()
        self.checkDefaults()
        self.checkLevelLimits()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_06():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 6: interpolate f(x,y) = exp(-5 * x^2) * cos(z), using the rleja rule")
    print("           employs adaptive construction\n")

    iNumInputs = 2
    iNumOutputs = 1
    iNumSamplesPerBatch = 1 # the model is set for a single sample per-batch
    def model(aX):
        # note that the model has to return a 2-D numpy.ndarray
        return np.ones((1,1)) * np.exp(-5.0 * aX[0, 0] ** 2.0) * np.cos(aX[0, 1])

    iTestGridSize = 33
    dx = np.linspace(-1.0, 1.0, iTestGridSize) # sample on a uniform grid
    aMeshX, aMeshY = np.meshgrid(dx, dx)
    aTestPoints = np.column_stack([aMeshX.reshape((iTestGridSize**2, 1)),
                                   aMeshY.reshape((iTestGridSize**2, 1))])
    aReferenceValues = np.exp(-5.0 * aTestPoints[:,0]**2) * np.cos(aTestPoints[:,1])
    aReferenceValues = aReferenceValues.reshape((aReferenceValues.shape[0], 1))

    def testGrid(grid, aTestPoints, aReferenceValues):
        aResult = grid.evaluateBatch(aTestPoints)
        for i in range(20):
            aX = aTestPoints[i,:]
        return np.max(np.abs(aResult - aReferenceValues))

    grid = Tasmanian.SparseGrid()
    grid.makeSequenceGrid(iNumInputs, iNumOutputs, 2, "level", "rleja")

    iNumThreads = 1

    print("{0:>8s}{1:>14s}".format("points", "error"))
    for i in range(4):
        iBudget = 50 * (i + 1)

        Tasmanian.constructAnisotropicSurrogate(lambda x, tid : model(x),
                                                iBudget, iNumThreads, iNumSamplesPerBatch, grid,
                                                "iptotal", 0)

        print("{0:>8d}{1:>14s}".format(grid.getNumPoints(),
              "{0:1.4e}".format(testGrid(grid, aTestPoints, aReferenceValues))))


if (__name__ == "__main__"):
    example_06()



import unittest
import TasmanianSG
import sys, os
import numpy as np

from random import uniform

import testConfigureData as tdata # needed to keep track of configured acceleration methods
import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Test the acceleration options:
    * consistency between CMake options and options reported by C++/Python
    * ability to switch between GPUs, read properties set GPUID, etc.
    * consistency between accelerated and reference evaluations
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkMeta(self):
        '''
        Check if the CMake options are propagated to the C++ library and the
        correct acceleration options are passed between C++ and Python.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        if (tdata.bEnableSyncTests):
            self.assertTrue((grid.isAccelerationAvailable("cpu-blas") == tdata.bHasBlas), "failed to match blas")
            self.assertTrue((grid.isAccelerationAvailable("gpu-cublas") == tdata.bHasCuBlas), "failed to match cublas")
            self.assertTrue((grid.isAccelerationAvailable("gpu-cuda") == tdata.bHasCuda), "failed to match cuda")
            self.assertTrue((grid.isAccelerationAvailable("gpu-default") == (tdata.bHasCuBlas or tdata.bHasCuda)), "failed to match cuda")

            # acceleration meta-data
            lsAvailableAcc = []
            if (tdata.bHasBlas): lsAvailableAcc.append("cpu-blas")
            if (tdata.bHasCuBlas): lsAvailableAcc.append("gpu-cublas")
            if (tdata.bHasCuda): lsAvailableAcc.append("gpu-cuda")
            grid.makeLocalPolynomialGrid(2, 1, 2, 1, 'semi-localp')
            for accel in lsAvailableAcc:
                grid.enableAcceleration(accel)
                sA = grid.getAccelerationType()
                bTest = ((accel not in sA) or (sA not in accel))
                self.assertFalse(bTest, "set/get Acceleration")

            lsAvailableAcc = []
            if (tdata.bHasCuBlas): lsAvailableAcc.append(("gpu-rocblas", "gpu-cublas"))
            if (tdata.bHasCuda): lsAvailableAcc.append(("gpu-hip", "gpu-cuda"))
            for accel in lsAvailableAcc:
                grid.enableAcceleration(accel[0])
                sA = grid.getAccelerationType()
                bTest = ((accel[1] not in sA) or (sA not in accel[1]))
                self.assertFalse(bTest, "set/get Acceleration with HIP - CUDA aliases")


    def checkMultiGPU(self):
        '''
        Check setting and resetting the GPU ID and reading the device names.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        self.assertTrue((grid.getGPUID() == 0), "did not default to gpu 0")

        if (grid.getNumGPUs() > 1):
            grid.setGPUID(1)
            self.assertTrue((grid.getGPUID() == 1), "did not set to gpu 1")
            if (not tdata.bUsingMSVC):
                sName = grid.getGPUName(1) # mostly checks for memory leaks and crashes

        if (grid.getNumGPUs() > 0):
            grid.setGPUID(0)
            self.assertTrue((grid.getGPUID() == 0), "did not set to gpu 0")
            if (not tdata.bUsingMSVC):
                sName = grid.getGPUName(0) # mostly checks for memory leaks and crashes

    def checkEvaluateConsistency(self):
        '''
        Check for consistency between accelerated and reference evaluations.
        In short, set a grid, load points, evaluate in different ways.
        Test all visible GPUs and combinations cuBlas/MAGMA, etc.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        aTestPointsCanonical = np.array([[ uniform(-1.0, 1.0) for j in range(2) ] for i in range(100) ])
        aTestPointsTransformed = np.array([[ uniform(3.0, 5.0) for j in range(2) ] for i in range(100) ])
        aDomainTransform = np.array([[3.0, 5.0],[3.0, 5.0]])

        iFastEvalSubtest = 6

        lTests = [ 'grid.makeGlobalGrid(2, 2, 4, "level", "clenshaw-curtis")',
                   'grid.makeGlobalGrid(2, 2, 4, "level", "chebyshev")',
                   'grid.makeSequenceGrid(2, 2, 4, "level", "leja")',
                   'grid.makeLocalPolynomialGrid(2, 3, 4, 1, "localp")',
                   'grid.makeLocalPolynomialGrid(2, 2, 4, 2, "localp")',
                   'grid.makeLocalPolynomialGrid(2, 1, 4, 3, "localp")',
                   'grid.makeLocalPolynomialGrid(2, 1, 4, 4, "semi-localp")',
                   'grid.makeWaveletGrid(2, 1, 3, 1)',
                   'grid.makeWaveletGrid(2, 1, 3, 3)',
                   'grid.makeFourierGrid(2, 1, 3, "level")' ]
        lTests = ['grid.makeLocalPolynomialGrid(2, 3, 4, 1, "localp")']

        iNumGPUs = grid.getNumGPUs()
        lsAccelTypes = ["none", "cpu-blas", "gpu-cuda", "gpu-cublas", "gpu-magma"]

        for sTest in lTests:
            for iI in range(2):
                iC = 0
                iGPU = 0
                if (tdata.iGPUID > -1):
                    iGPU = tdata.iGPUID

                while (iC < len(lsAccelTypes)):
                    sAcc = lsAccelTypes[iC]
                    exec(sTest)
                    if (iI == 0):
                        aTestPoints = aTestPointsCanonical
                    else:
                        aTestPoints = aTestPointsTransformed
                        grid.setDomainTransform(aDomainTransform)

                    grid.enableAcceleration(sAcc)
                    if ((sAcc == "gpu-cuda") or (sAcc == "gpu-cublas") or (sAcc == "gpu-magma")):
                        if (iGPU < grid.getNumGPUs()): # without cuda or cublas, NumGPUs is 0 and cannot set GPU
                            grid.setGPUID(iGPU)

                    ttc.loadExpN2(grid)

                    aRegular = np.array([grid.evaluateThreadSafe(aTestPoints[i,:]) for i in range(aTestPoints.shape[0]) ])
                    aBatched = grid.evaluateBatch(aTestPoints)
                    np.testing.assert_almost_equal(aRegular, aBatched, 14, "Batch evaluation test not equal: {0:1s}, acceleration: {1:1s}, gpu: {2:1d}".format(sTest, sAcc, iGPU), True)

                    aFast = np.array([ grid.evaluate(aTestPoints[i,:]) for i in range(iFastEvalSubtest) ])
                    np.testing.assert_almost_equal(aRegular[0:iFastEvalSubtest,:], aFast, 14, "Batch evaluation test not equal: {0:1s}, acceleration: {1:1s}, gpu: {2:1d}".format(sTest, sAcc, iGPU), True)

                    if ((sAcc == "gpu-cuda") or (sAcc == "gpu-cublas") or (sAcc == "gpu-magma")):
                        if (tdata.iGPUID == -1):
                            iGPU += 1
                            if (iGPU >= iNumGPUs):
                                iC += 1
                                iGPU = 0
                        else:
                            iC += 1
                    else:
                        iC += 1

    def performAccelerationTest(self):
        self.checkMeta()
        self.checkMultiGPU()
        self.checkEvaluateConsistency()



#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import DREAM
from Tasmanian import SparseGrid
from Tasmanian import makeSequenceGrid
from Tasmanian import loadNeededPoints
import numpy

def example_03():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 3: set the inference problem: identify x_0 and x_1 model parameters")
    print("           from data (noise free example)")
    print("           model: f(x) = sin(x_0*M_PI*t + x_1),")
    print("           data: d = sin(5*M_PI*t + 0.3*M_PI) + sin(10*M_PI*t + 0.1*M_PI)")
    print("           compared to Example 2, the data is a superposition of two signals")
    print("           and the posterior is multi-modal")
    print("             -- problem setup --")
    print("           t in [0,1], t is discretized with 32 equidistant nodes")
    print("           the likelihood is exp(- 16 * (f(x) - d)^2)")
    print("           using a sparse grid to interpolate the model\n")

    iNumDimensions = 2
    iNumOutputs = 32
    iNumChains = 500
    iNumBurnupIterations = 1000
    iNumCollectIterations = 300

    def model(aX):
        # using 32 nodes in the interior of (0.0, 1.0)
        t = numpy.linspace(1.0 / 64.0, 1.0 - 1.0 / 64.0, 32)
        return numpy.sin(numpy.pi * aX[0] * t + aX[1])

    aData = model([5.0, 0.3 * numpy.pi]) + model([10.0, 0.1 * numpy.pi])

    grid = makeSequenceGrid(iNumDimensions, iNumOutputs, 30, 'iptotal', 'leja')

    domain_lower = numpy.array([ 1.0, -0.1])
    domain_upper = numpy.array([12.0,  1.7])
    grid.setDomainTransform(numpy.column_stack([domain_lower, domain_upper]))

    loadNeededPoints(lambda x, tid : model(x), grid, 2)

    likely = DREAM.LikelihoodGaussIsotropic(1.0 / float(iNumOutputs), aData)

    state = DREAM.State(iNumChains, grid)
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 DREAM.Posterior(grid, likely, prior = "uniform", typeForm = DREAM.typeLogform),
                 DREAM.Domain(grid),
                 state,
                 DREAM.IndependentUpdate("gaussian", 0.01),
                 DREAM.DifferentialUpdate(90),
                 typeForm = DREAM.typeLogform)

    aSamples = state.getHistory()

    # split the frequencies into low and high for the two modes of the posterior
    # here 6.5 is the mid-point of the domain
    low_frequency = numpy.average(numpy.array(
        [aSamples[i,0] for i in range(aSamples.shape[0]) if aSamples[i,0] < 6.5]
    ))
    low_correction = numpy.average(numpy.array(
        [aSamples[i,1] for i in range(aSamples.shape[0]) if aSamples[i,0] < 6.5]
    ))
    high_frequency = numpy.average(numpy.array(
        [aSamples[i,0] for i in range(aSamples.shape[0]) if aSamples[i,0] >= 6.5]
    ))
    high_correction = numpy.average(numpy.array(
        [aSamples[i,1] for i in range(aSamples.shape[0]) if aSamples[i,0] >= 6.5]
    ))

    print("Inferred values:")
    print(" low   frequency:{0:13.6f}   error:{1:14.6e}".format(low_frequency,
                                                     numpy.abs(low_frequency - 5.0)))
    print(" low  correction:{0:13.6f}   error:{1:14.6e}\n".format(low_correction,
                                                     numpy.abs(low_correction - 0.3 * numpy.pi)))
    print(" high  frequency:{0:13.6f}   error:{1:14.6e}".format(high_frequency,
                                                     numpy.abs(high_frequency - 10.0)))
    print(" high correction:{0:13.6f}   error:{1:14.6e}".format(high_correction,
                                                    numpy.abs(high_correction - 0.1 * numpy.pi)))

if __name__ == "__main__":
    example_03()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_05():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 5: interpolate f(x,y) = exp(-x^2) * cos(y), using leja rule")
    print("           employ adaptive refinement to increase accuracy per samples")

    iNumInputs = 2
    iNumOutputs = 1
    def model(aX):
        return np.ones((1,)) * np.exp(-aX[0] * aX[0]) * np.cos(aX[1])

    iTestGridSize = 33
    dx = np.linspace(-1.0, 1.0, iTestGridSize) # sample on a uniform grid
    aMeshX, aMeshY = np.meshgrid(dx, dx)
    aTestPoints = np.column_stack([aMeshX.reshape((iTestGridSize**2, 1)),
                                   aMeshY.reshape((iTestGridSize**2, 1))])
    aReferenceValues = np.exp(-aTestPoints[:,0]**2) * np.cos(aTestPoints[:,1])
    aReferenceValues = aReferenceValues.reshape((aReferenceValues.shape[0], 1))

    def testGrid(grid, aTestPoints, aReferenceValues):
        aResult = grid.evaluateBatch(aTestPoints)
        for i in range(20):
            aX = aTestPoints[i,:]
        return np.max(np.abs(aResult - aReferenceValues))


    iInitialLevel = 5

    grid_isotropic = Tasmanian.SparseGrid()
    grid_iptotal   = Tasmanian.SparseGrid()
    grid_icurved   = Tasmanian.SparseGrid()
    grid_surplus   = Tasmanian.SparseGrid()
    grid_isotropic.makeGlobalGrid(iNumInputs, iNumOutputs, iInitialLevel, "level", "leja")
    grid_iptotal.copyGrid(grid_isotropic)
    grid_icurved.copyGrid(grid_isotropic)
    grid_surplus.copyGrid(grid_isotropic)

    iNumThreads = 1
    iBudget = 100
    print("{0:>22s}{1:>22s}{2:>22s}{3:>22s}".format("isotropic", "iptotal", "ipcurved", "surplus"))
    print("{0:>8s}{1:>14s}{0:>8s}{1:>14s}{0:>8s}{1:>14s}{0:>8s}{1:>14s}".format("points", "error"))

    bBelowBudget = True
    while(bBelowBudget):
        sInfo = ""
        if (grid_isotropic.getNumLoaded() < iBudget):
            Tasmanian.loadNeededValues(lambda x, tid : model(x), grid_isotropic, iNumThreads)
            sInfo += "{0:>8d}{1:>14s}".format(grid_isotropic.getNumLoaded(),
                     "{0:1.4e}".format(testGrid(grid_isotropic, aTestPoints, aReferenceValues)))

            iLevel = 0
            while(grid_isotropic.getNumNeeded() == 0):
                grid_isotropic.updateGlobalGrid(iLevel, "level")
                iLevel += 1
        else:
            sInfo += "{0:>22s}".format("")

        if (grid_iptotal.getNumLoaded() < iBudget):
            Tasmanian.loadNeededValues(lambda x, tid : model(x), grid_iptotal, iNumThreads)
            sInfo += "{0:>8d}{1:>14s}".format(grid_iptotal.getNumLoaded(),
                     "{0:1.4e}".format(testGrid(grid_iptotal, aTestPoints, aReferenceValues)))

            grid_iptotal.setAnisotropicRefinement("iptotal", 10, 0);
        else:
            sInfo += "{0:>22s}".format("")

        if (grid_icurved.getNumLoaded() < iBudget):
            Tasmanian.loadNeededValues(lambda x, tid : model(x), grid_icurved, iNumThreads)
            sInfo += "{0:>8d}{1:>14s}".format(grid_icurved.getNumLoaded(),
                     "{0:1.4e}".format(testGrid(grid_icurved, aTestPoints, aReferenceValues)))

            grid_icurved.setAnisotropicRefinement("ipcurved", 10, 0);
        else:
            sInfo += "{0:>22s}".format("")

        if (grid_surplus.getNumLoaded() < iBudget):
            Tasmanian.loadNeededValues(lambda x, tid : model(x), grid_surplus, iNumThreads)
            sInfo += "{0:>8d}{1:>14s}".format(grid_surplus.getNumLoaded(),
                     "{0:1.4e}".format(testGrid(grid_surplus, aTestPoints, aReferenceValues)))

            grid_surplus.setSurplusRefinement(1.E-8, 0)
        else:
            sInfo += "{0:>22s}".format("")

        print(sInfo)
        bBelowBudget = (grid_isotropic.getNumLoaded() < iBudget
                       or grid_icurved.getNumLoaded() < iBudget
                       or grid_icurved.getNumLoaded() < iBudget
                       or grid_surplus.getNumLoaded() < iBudget)


if (__name__ == "__main__"):
    example_05()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from ctypes import c_char_p, c_int, c_double, c_void_p, POINTER, cdll, create_string_buffer
import numpy as np
import sys

from TasmanianConfig import __path_libdream__
from TasmanianConfig import TasmanianInputError as InputError

from TasmanianSG import TasmanianSparseGrid as SparseGrid

pLibDTSG = cdll.LoadLibrary(__path_libdream__)

pLibDTSG.tsgMakeDreamState.restype = c_void_p
pLibDTSG.tsgMakeDreamState.argtypes = [c_int, c_int]

pLibDTSG.tsgDeleteDreamState.argtypes = [c_void_p]

pLibDTSG.tsgDreamStateGetDims.restype = c_int
pLibDTSG.tsgDreamStateGetDims.argtypes = [c_void_p]

pLibDTSG.tsgDreamStateGetChains.restype = c_int
pLibDTSG.tsgDreamStateGetChains.argtypes = [c_void_p]

pLibDTSG.tsgDreamStateGetNumHistory.restype = c_int
pLibDTSG.tsgDreamStateGetNumHistory.argtypes = [c_void_p]

pLibDTSG.tsgDreamStateSet.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgDreamStateGetHistory.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgDreamStateGetHistoryPDF.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgDreamStateGetMeanVar.argtypes = [c_void_p, POINTER(c_double), POINTER(c_double)]
pLibDTSG.tsgDreamStateGetMode.argtypes = [c_void_p, POINTER(c_double)]

pLibDTSG.tsgDreamStateGetRate.restype = c_double
pLibDTSG.tsgDreamStateGetRate.argtypes = [c_void_p]

class DreamState:
    '''
    Wrapper to class TasDREAM::TasmanianDREAM
    '''
    def __init__(self, iNumChains, GridOrIntDimensions):
        '''
        Constructs a new state with given number of chains and dimensions.
        The dimensions can be given as an integer or inferred from
        a Tasmanian.SparseGrid object.

        iNumChains: positive integer indicating the number of chains.
        GridOrIntDimensions: either a positive integer or a non-empty
            Tasmanian.SparseGrid object that gives the dimensions
        '''
        self.TasmanainDreamState = True
        if (isinstance(GridOrIntDimensions, SparseGrid)):
            self.pStatePntr = pLibDTSG.tsgMakeDreamState(iNumChains, GridOrIntDimensions.getNumDimensions())
        else:
            self.pStatePntr = pLibDTSG.tsgMakeDreamState(iNumChains, GridOrIntDimensions)

    def __del__(self):
        '''
        Deletes an instance of the DREAM state.
        '''
        pLibDTSG.tsgDeleteDreamState(self.pStatePntr)

    def getNumDimensions(self):
        '''
        Return the number of dimensions.
        '''
        return pLibDTSG.tsgDreamStateGetDims(self.pStatePntr)

    def getNumChains(self):
        '''
        Return the number of chains.
        '''
        return pLibDTSG.tsgDreamStateGetChains(self.pStatePntr)

    def getNumHistory(self):
        '''
        Return the number of samples saved in the history.
        '''
        return pLibDTSG.tsgDreamStateGetNumHistory(self.pStatePntr)

    def setState(self, llfNewState):
        '''
        Set a new state for the DREAM chains.

        llfNewState: is a two dimensional numpy.ndarray with
            .shape[0] = .getNumChains()
            .shape[1] = .getNumDimensions()
        '''
        iNumChains = self.getNumChains()
        iNumDims = self.getNumDimensions()
        if (llfNewState.shape[0] != iNumChains):
            raise InputError("llfNewState", "llfNewState.shape[0] should match the number of chains")
        if (llfNewState.shape[1] != iNumDims):
            raise InputError("llfNewState", "llfNewState.shape[1] should match the number of dimensions")
        pLibDTSG.tsgDreamStateSet(self.pStatePntr,
                                  np.ctypeslib.as_ctypes(llfNewState.reshape((iNumChains * iNumDims,))))

    def getHistory(self):
        '''
        Returns the saved history in a two dimensional numpy.ndarray
        '''
        iNumDims = self.getNumDimensions()
        iNumHistory = self.getNumHistory()
        aResult = np.empty((iNumDims * iNumHistory,), np.float64)
        pLibDTSG.tsgDreamStateGetHistory(self.pStatePntr, np.ctypeslib.as_ctypes(aResult))
        return aResult.reshape((iNumHistory, iNumDims))

    def getHistoryPDF(self):
        '''
        Returns the saved pdf in a one dimensional numpy.ndarray
        '''
        iNumHistory = self.getNumHistory()
        aResult = np.empty((NumHistory,), np.float64)
        pLibDTSG.tsgDreamStateGetHistoryPDF(self.pStatePntr, np.ctypeslib.as_ctypes(aResult))
        return aResult

    def getHistoryMeanVariance(self):
        '''
        Returns two one dimensional numpy.ndarrays corresponding
        to the mean and variance of the history.
        '''
        iNumDims = self.getNumDimensions()
        aMean = np.empty((iNumDims,))
        aVar  = np.empty((iNumDims,))
        pLibDTSG.tsgDreamStateGetMeanVar(self.pStatePntr,
                                         np.ctypeslib.as_ctypes(aMean),
                                         np.ctypeslib.as_ctypes(aVar))
        return aMean, aVar

    def getApproximateMode(self):
        '''
        Returns the approximate mode, i.e., the sample with highest probability density.
        '''
        iNumDims = self.getNumDimensions()
        aMode = np.empty((iNumDims,))
        pLibDTSG.tsgDreamStateGetMode(self.pStatePntr, np.ctypeslib.as_ctypes(aMode))
        return aMode

    def getAcceptanceRate(self):
        '''
        Return the acceptance rate accumulated with the history.
        '''
        return pLibDTSG.tsgDreamStateGetRate(self.pStatePntr)




import sys, site

# do standard skbuild setup
from packaging.version import LegacyVersion
from skbuild.exceptions import SKBuildError
from skbuild.cmaker import get_cmake_version
from skbuild import setup  # This line replaces 'from setuptools import setup'

# Add CMake as a build requirement if cmake is not installed or too old
setup_requires = []
try:
    if LegacyVersion(get_cmake_version()) < LegacyVersion("3.10"):
        setup_requires.append('cmake>=3.10')
except SKBuildError:
    setup_requires.append('cmake>=3.10')
setup_requires.append('numpy>=1.10')

with open('README.md', 'r') as fh:
     readme_file = fh.readlines()

long_description = ""
for line in readme_file[3:]:
    if line.rstrip() == "Quick Install":
        break
    else:
        long_description += line

long_description += "### Quick Install\n Tasmanian supports `--user` and venv install only, see the on-line documentation for details.\n"

# find out whether this is a virtual environment, real_prefix is an older test, base_refix is the newer one
if hasattr(sys, 'real_prefix') or (hasattr(sys, 'base_prefix') and sys.base_prefix != sys.prefix):
    final_install_path = sys.prefix # sys.prefix points to the virtual environment root
    isvirtual = True
else:
    isvirtual = False
    try:
        final_install_path = site.getuserbase()
    except:
        import os
        # some implementations do not provide compatible 'site' package, assume default Linux behavior
        final_install_path = os.getenv('HOME') + "/.local/"

# check if using OSX Framework environment
isosxframework = False
if sys.platform == 'darwin':
    try:
        if 'python/site-packages' in site.getusersitepackages():
            # appears to be Mac Framework using Library/Python/X.Y/lib/python/site-packages
            isosxframework = True
    except:
        # cannot determine if using Mac Framework
        pass

# setup cmake arguments
cmake_args=[
        '-DCMAKE_BUILD_TYPE=Release',
        '-DBUILD_SHARED_LIBS=ON',
        '-DTasmanian_ENABLE_RECOMMENDED:BOOL=ON',
        '-DTasmanian_ENABLE_PYTHON:BOOL=ON',
        '-DPYTHON_EXECUTABLE:PATH={0:1s}'.format(sys.executable),
        '-DTasmanian_python_pip_final:PATH={0:1s}/'.format(final_install_path),
        ]
if isvirtual:
    cmake_args.append('-DTasmanian_windows_virtual:BOOL=ON')
if isosxframework:
    cmake_args.append('-DTasmanian_osx_framework:BOOL=ON')


# call the actual package setup command
setup(
    name='Tasmanian',
    version='7.7',
    author='Miroslav Stoyanov',
    author_email='stoyanovmk@ornl.gov',
    description='UQ library for sparse grids and Bayesian inference',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://tasmanian.ornl.gov',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: C++',
        'Operating System :: OS Independent',
    ],
    install_requires=['numpy>=1.10'],
    ### cmake portion of the setup, specific to skbuild ###
    setup_requires=setup_requires,
    cmake_args=cmake_args
)



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_04():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 4: interpolate f(x,y) = exp(-x^2) * cos(y), using clenshaw-curtis iptotal rule")

    iNumInputs = 2
    iNumOutputs = 1

    aPointOfInterest = np.array([0.3, 0.7])
    aReferenceSolution = np.exp(-aPointOfInterest[0]**2) * np.cos(aPointOfInterest[1])

    for prec in [6, 12]:
        grid = Tasmanian.makeGlobalGrid(iNumInputs, iNumOutputs, prec,
                                        "iptotal", "clenshaw-curtis")
        aPoints = grid.getNeededPoints()

        aModelValues = np.exp(-aPoints[:,0]**2) * np.cos(aPoints[:,1])
        grid.loadNeededValues(aModelValues.reshape((aModelValues.shape[0], iNumOutputs)))

        # when using multiple points at once, evaluateBatch() is more efficient
        aResult = grid.evaluate(aPointOfInterest)
        fError = np.abs(aResult[0] - aReferenceSolution)

        print("\n    using total degree polynomials: {0:>2d}th degree".format(prec))
        print("                      the grid has: {0:1d} points".format(grid.getNumPoints()))
        print("          interpolant at (0.3,0.7): {0:1.5e}".format(aResult[0]))
        print("                             error: {0:1.5e}".format(fError))


if (__name__ == "__main__"):
    example_04()



import unittest
import TasmanianSG
import sys, os
import numpy as np

import testConfigureData as tdata # needed for Gauss-Patterson table file
import testCommon
ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Test the read/write functionality in different formats, also the
    reading of various meta-data (i.e., version or gpu info)
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkMetaIO(self):
        '''
        Test the version I/O, available acceleration, gpu info, print stats
        '''
        grid = TasmanianSG.TasmanianSparseGrid()
        print("Tasmanian Sparse Grids version: {0:1s}".format(grid.getVersion()))
        print("                 version major: {0:1d}".format(grid.getVersionMajor()))
        print("                 version minor: {0:1d}".format(grid.getVersionMinor()))
        print("                       License: {0:1s}".format(grid.getLicense()))
        if (grid.isOpenMPEnabled()):
            sStatus = "Enabled"
        else:
            sStatus = "Disabled"
        print("                        OpenMP: {0:1s}".format(sStatus))
        sAvailableAcceleration = ""
        sGPUbackend = "none"
        if (grid.isCudaEnabled()): sGPUbackend = "CUDA"
        if (grid.isHipEnabled()): sGPUbackend = "ROCm/HIP"
        if (grid.isDpcppEnabled()): sGPUbackend = "oneAPI/DPC++"
        print("                   GPU backend: {0:1s}".format(sGPUbackend))
        for s in ["cpu-blas", "gpu-cublas", "gpu-cuda", "gpu-magma"]:
            if (grid.isAccelerationAvailable(s)):
                sAvailableAcceleration += " " + s
        print("        Available acceleration:{0:1s}".format(sAvailableAcceleration))

        print("                Available GPUs:")
        if (grid.getNumGPUs() > 0):
            for iGPU in range(grid.getNumGPUs()):
                sName = ""
                if ("@CMAKE_CXX_COMPILER_ID@" == "MSVC"):
                    sName = "Unavailable under Windows"
                else:
                    sName = grid.getGPUName(iGPU)
                sMem = grid.getGPUMemory(iGPU)
                print("     {0:2d}: {1:20s} with{2:6d}MB RAM".format(iGPU, sName, sMem))
        else:
            print("            none")

        # covers the printStats() in python and C++
        grid.printStats() # empty grid
        grid.makeGlobalGrid(2, 0, 1, "level", "gauss-gegenbauer", [], 3.0)
        grid.printStats()
        grid.makeGlobalGrid(2, 0, 1, "level", "gauss-jacobi", [], 3.0, 3.0)
        grid.printStats()
        grid.makeGlobalGrid(2, 1, 1, "level", "custom-tabulated", [], 0.0, 0.0, tdata.sGaussPattersonTableFile)
        grid.printStats()
        grid.makeSequenceGrid(2, 1, 3, "level", "rleja")
        ttc.loadExpN2(grid)
        grid.printStats()
        grid.makeLocalPolynomialGrid(1, 1, 3)
        grid.setDomainTransform(np.array([[2.0, 3.0]]))
        grid.printStats()
        grid.makeWaveletGrid(2, 1, 1)
        grid.printStats()
        grid.makeFourierGrid(3, 1, 1, "level")
        grid.enableAcceleration("gpu-cuda")
        grid.printStats()

    def checkReadWriteGlobal(self):
        '''
        Test reading and writing of Global grids.
        '''
        # iDimension, iOutputs, iDepth, sType, sRule, fAlpha, fBeta, useTransform, loadFunciton, limitLevels
        lGrids = [[3, 2, 2, "level", "leja", 0.0, 0.0, False, False, False],
                  [2, 1, 4, "level", "clenshaw-curtis", 0.0, 0.0, False, False, False],
                  [3, 1, 3, "level", "rleja", 0.0, 0.0, True, False, False],
                  [3, 1, 2, "iptotal", "chebyshev", 0.0, 0.0, False, True, False],
                  [3, 1, 3, "level", "leja", 0.0, 0.0, True, True, False],
                  [3, 1, 3, "level", "leja", 0.0, 0.0, True, True, True],
                  [2, 1, 5, "qptotal", "gauss-hermite", 1.0, 3.0, False, False, False],
                  [3, 1, 2, "level", "gauss-laguerre", 3.0, 0.0, False, False, True],]
        aTransform = np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]])

        for lT in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            if (lT[7]):
                gridA.makeGlobalGrid(lT[0], lT[1], lT[2], lT[3], lT[4], [], lT[5], lT[6], [], [3, 2, 1])
            else:
                gridA.makeGlobalGrid(lT[0], lT[1], lT[2], lT[3], lT[4], [], lT[5], lT[6])
            #gridA.printStats()
            if (lT[7]):
                gridA.setDomainTransform(aTransform)
            if (lT[8]):
                ttc.loadExpN2(gridA)

            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeGlobalGrid(1, 0, 1, "level", "rleja")
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

        # test an error message from wrong read
        try:
            gridB.read("Test_If_Bogus_Filename_Produces_an_Error")
        except TasmanianSG.TasmanianInputError as TSGError:
            TSGError.bShowOnExit = False
            self.assertEqual(TSGError.sVariable, "sFilename", "Reading a bogus file properly failed, but the error information is wrong.")

        # custom rule test
        gridA = TasmanianSG.TasmanianSparseGrid()
        gridB = TasmanianSG.TasmanianSparseGrid()
        gridA.makeGlobalGrid(2, 0, 4, 'level', 'custom-tabulated', [], 0.0, 0.0, tdata.sGaussPattersonTableFile)
        gridB = TasmanianSG.TasmanianSparseGrid()
        gridB.makeGlobalGrid(2, 0, 4, 'level', 'gauss-patterson')
        ttc.compareGrids(gridA, gridB, bTestRuleNames = False)
        gridA.write("testSave", bUseBinaryFormat = False)
        gridB.makeGlobalGrid(2, 0, 4, 'level', 'clenshaw-curtis')
        gridB.read("testSave")
        ttc.compareGrids(gridA, gridB)
        gridA.write("testSave", bUseBinaryFormat = True)
        gridB.makeGlobalGrid(3, 0, 4, 'level', 'leja')
        gridB.read("testSave")
        ttc.compareGrids(gridA, gridB)

    def checkReadWriteSequence(self):
        '''
        Test reading and writing of Sequence grids.
        '''
        # iDimension, iOutputs, iDepth, sType, sRule, useTransform, loadFunciton, limitLevels
        lGrids = [[3, 2, 2, "level", "leja", False, False, False],
                  [2, 1, 4, "level", "max-lebesgue", False, False, False],
                  [3, 1, 3, "level", "rleja", True, False, False],
                  [3, 1, 3, "level", "rleja", True, False, True],
                  [3, 1, 2, "iptotal", "min-delta", False, True, False],
                  [3, 1, 3, "level", "leja", True, True, False],
                  [3, 1, 3, "level", "leja", True, True, True],]

        for lT in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            if (lT[7]):
                gridA.makeSequenceGrid(lT[0], lT[1], lT[2], lT[3], lT[4], [], [2, 3, 1])
            else:
                gridA.makeSequenceGrid(lT[0], lT[1], lT[2], lT[3], lT[4])
            if (lT[5]):
                gridA.setDomainTransform(np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]]))
            if (lT[6]):
                ttc.loadExpN2(gridA)

            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeGlobalGrid(1, 0, 1, "level", "rleja")
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

    def checkReadWriteLocalp(self):
        '''
        Test reading and writing of Localp grids.
        '''
        # iDimension, iOutputs, iDepth, iorder, sRule, useTransform, loadFunciton, limitLevels
        lGrids = [[3, 2, 2, 0, "localp", False, False, False],
                  [3, 0, 2, 0, "localp-boundary", False, False, False],
                  [2, 1, 4, 1, "semi-localp", False, False, False],
                  [3, 1, 3, 2, "localp", True, False, False],
                  [3, 1, 2, 3, "localp-zero", False, True, False],
                  [3, 1, 2, 3, "localp-zero", False, True, True],
                  [3, 1, 3, 4, "semi-localp", True, True, False],
                  [3, 1, 3, -1, "semi-localp", True, True, False],
                  [3, 1, 3, -1, "semi-localp", True, True, True],]

        for lT in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            if (lT[7]):
                gridA.makeLocalPolynomialGrid(lT[0], lT[1], lT[2], lT[3], lT[4], [3, 1, 2])
            else:
                gridA.makeLocalPolynomialGrid(lT[0], lT[1], lT[2], lT[3], lT[4])
            if (lT[5]):
                gridA.setDomainTransform(np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]]))
            if (lT[6]):
                ttc.loadExpN2(gridA)

            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeGlobalGrid(1, 0, 1, "level", "rleja")
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

    def checkReadWriteWavelet(self):
        '''
        Test reading and writing of Wavelet grids.
        '''
        # iDimension, iOutputs, iDepth, iOrder, useTransform, loadFunciton
        lGrids = [[3, 2, 2, 1, False, False, False],
                  [3, 0, 2, 1, False, False, False],
                  [2, 1, 4, 1, False, False, False],
                  [3, 1, 1, 3, True, False, False],
                  [3, 1, 1, 3, True, False, True],
                  [3, 1, 2, 1, False, True, False],
                  [3, 1, 2, 3, True, True, True],
                  [3, 1, 2, 3, True, True, False],]

        for lT in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            if (lT[6]):
                gridA.makeWaveletGrid(lT[0], lT[1], lT[2], lT[3], [1, 1, 2])
            else:
                gridA.makeWaveletGrid(lT[0], lT[1], lT[2], lT[3])
            if (lT[4]):
                gridA.setDomainTransform(np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]]))
            if (lT[5]):
                ttc.loadExpN2(gridA)

            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeGlobalGrid(1, 0, 1, "level", "rleja")
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

    def checkReadWriteFourier(self):
        '''
        Test reading and writing of Fourier grids.
        '''
        # iDimension, iOutputs, iDepth, useTransform, loadFunction, useLevelLimits
        lGrids = [[3, 2, 2, False, False, False],
                  [2, 1, 4, False, False, False],
                  [3, 1, 1, True, False, False],
                  [3, 1, 1, True, False, True],
                  [3, 1, 2, False, True, False],
                  [3, 1, 2, True, True, True],
                  [3, 1, 2, True, True, False],]

        for lT in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            if (lT[5]):
                gridA.makeFourierGrid(lT[0], lT[1], lT[2], 'level', [1, 1, 2])
            else:
                gridA.makeFourierGrid(lT[0], lT[1], lT[2], 'level')
            if (lT[3]):
                gridA.setDomainTransform(np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]]))
            if (lT[4]):
                ttc.loadExpN2(gridA)

            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeGlobalGrid(1, 0, 1, "level", "rleja")
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

    def checkCopySubgrid(self):
        grid = TasmanianSG.TasmanianSparseGrid() # test grid
        gridTotal = TasmanianSG.TasmanianSparseGrid() # test grid
        gridRef1 = TasmanianSG.TasmanianSparseGrid() # test grid
        gridRef2 = TasmanianSG.TasmanianSparseGrid() # test grid
        gridRef3 = TasmanianSG.TasmanianSparseGrid() # test grid
        gridRef4 = TasmanianSG.TasmanianSparseGrid() # test grid
        # outputs:   source       0         0, 1, 2     2, 3, 4       5
        lGrids = ['gridTotal', 'gridRef1', 'gridRef2', 'gridRef3', 'gridRef4']
        lOutputs = [6, 1, 3, 3, 1];
        lMake = ['.makeGlobalGrid(2, iOutputs, 4, "level", "clenshaw-curtis")',
                 '.makeSequenceGrid(2, iOutputs, 4, "level", "rleja")',
                 '.makeLocalPolynomialGrid(2, iOutputs, 4, 2)',
                 '.makeWaveletGrid(2, iOutputs, 3)',
                 '.makeFourierGrid(2, iOutputs, 4, "level")']

        for sMake in lMake:
            for iI in range(len(lOutputs)):
                iOutputs = lOutputs[iI]
                exec(lGrids[iI] + sMake)

            gridTotalValues = []
            gridRef1Values  = []
            gridRef2Values  = []
            gridRef3Values  = []
            gridRef4Values  = []

            aPoints = gridTotal.getPoints()
            for iI in range(aPoints.shape[0]):
                lModel = [i * np.exp(aPoints[iI][0] + aPoints[iI][1]) for i in range(1, 7)]
                gridTotalValues.append(lModel)
                gridRef1Values.append(lModel[0:1])
                gridRef2Values.append(lModel[0:3])
                gridRef3Values.append(lModel[2:5])
                gridRef4Values.append(lModel[5:6])

            for sGrid in lGrids:
                exec(sGrid + ".loadNeededPoints(np.row_stack(" + sGrid + "Values))")

            grid.copyGrid(gridTotal)
            ttc.compareGrids(grid, gridTotal)
            grid.copyGrid(gridTotal, 0, 1)
            ttc.compareGrids(grid, gridRef1)
            grid.copyGrid(gridTotal, 0, 3)
            ttc.compareGrids(grid, gridRef2)
            grid.copyGrid(gridTotal, 2, 5)
            ttc.compareGrids(grid, gridRef3)
            grid.copyGrid(gridTotal, 5, 6)
            ttc.compareGrids(grid, gridRef4)

    def checkReadWriteMisc(self):
        '''
        Test reading and writing of domain transforms and testing all rules.
        '''
        lGrids = ['gridA.makeGlobalGrid(3, 2, 4, "level", "clenshaw-curtis"); gridA.setDomainTransform(aTransform); gridA.setConformalTransformASIN(np.array([3,4,5]))',
                  'gridA.makeGlobalGrid(3, 2, 4, "level", "gauss-legendre"); gridA.setConformalTransformASIN(np.array([3,5,1]))',
                  'gridA.makeSequenceGrid(2, 2, 5, "level", "leja"); gridA.setConformalTransformASIN(np.array([0,4]))',
                  'gridA.makeLocalPolynomialGrid(3, 1, 4, 2, "localp"); gridA.setDomainTransform(aTransform); gridA.setConformalTransformASIN(np.array([5,3,0]))',
                  'gridA.getNumPoints()']
        aTransform = np.array([[0.0,1.0],[0.0,1.0],[-2.0,-1.0]])

        for sGrid in lGrids:
            gridA = TasmanianSG.TasmanianSparseGrid()
            gridB = TasmanianSG.TasmanianSparseGrid()

            exec(sGrid)
            gridA.write("testSave", bUseBinaryFormat = False)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridA.write("testSave", bUseBinaryFormat = True)
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.read("testSave")
            ttc.compareGrids(gridA, gridB)

            gridB.makeSequenceGrid(1, 1, 0, "level", "leja");
            gridB.makeLocalPolynomialGrid(1, 1, 0)
            gridB.copyGrid(gridA)
            ttc.compareGrids(gridA, gridB)

        # Make a grid with every possible rule (catches false-positive and memory crashes)
        for sType in TasmanianSG.lsTsgGlobalTypes:
            for sRule in TasmanianSG.lsTsgGlobalRules:
                if ("custom-tabulated" in sRule):
                    gridA.makeGlobalGrid(2, 0, 2, sType, sRule, sCustomFilename = tdata.sGaussPattersonTableFile)
                else:
                    gridA.makeGlobalGrid(2, 0, 2, sType, sRule)
                gridA.write("testSave", bUseBinaryFormat = False)
                gridB.read("testSave")
                ttc.compareGrids(gridA, gridB)
                gridB.makeGlobalGrid(1, 0, 0, "level", "clenshaw-curtis")
                gridA.write("testSave", bUseBinaryFormat = True)
                gridB.read("testSave")

        for sType in TasmanianSG.lsTsgGlobalTypes:
            for sRule in TasmanianSG.lsTsgSequenceRules:
                gridA.makeSequenceGrid(2, 1, 3, sType, sRule)
                gridA.write("testSave", bUseBinaryFormat = False)
                gridB.read("testSave")
                ttc.compareGrids(gridA, gridB)
                gridB.makeGlobalGrid(1, 0, 0, "level", "clenshaw-curtis")
                gridA.write("testSave", bUseBinaryFormat = True)
                gridB.read("testSave")
                ttc.compareGrids(gridA, gridB)

    def checkReadWriteCustomTabulated(self):
        '''
        Test reading and writing of the CustomTabulated class.
        '''
        description = "testCT"
        def create_nodes(j):
            return np.linspace(-1.0, 1.0, num=j)
        def create_weights(j):
            weights = np.linspace(0.0, 1.0, num=j)
            weights = 2.0 * weights / weights.sum()
            return weights
        # Read and write from explicitly given data.
        for i in range(4):
            num_levels = i
            num_nodes = np.array([3*(j+1) for j in range(i)])
            precision = np.array([2*(j+1)-1 for j in range(i)])
            nodes = [create_nodes(j) for j in num_nodes]
            weights = [create_weights(j) for j in num_nodes]
            ctA = TasmanianSG.makeCustomTabulatedFromData(num_levels, num_nodes, precision, nodes, weights, description)            # .
            ctB = TasmanianSG.CustomTabulated()
            ctA.write("testSave")
            ctB.read("testSave")
            for i in range(num_levels):
                read_weights, read_nodes = ctB.getWeightsNodes(i)
                np.testing.assert_almost_equal(read_weights, weights[i])
                np.testing.assert_almost_equal(read_nodes, nodes[i])
        # Read and write from a file.
        ctA = TasmanianSG.makeCustomTabulatedFromFile(tdata.sGaussPattersonTableFile)
        ctB = TasmanianSG.CustomTabulated()
        ctA.write("testSave")
        ctB.read("testSave")
        ttc.compareCustomTabulated(ctA, ctB)
        for i in range(ctB.getNumLevels()):
            read_weights, read_nodes = ctB.getWeightsNodes(i)
            grid = TasmanianSG.makeGlobalGrid(1, 0, i, "level", "gauss-patterson")
            np.testing.assert_almost_equal(read_weights, grid.getQuadratureWeights())
            np.testing.assert_almost_equal(read_nodes, grid.getPoints().flatten())
        # Test an error message from wrong read.
        try:
            ctB.read("Test_If_Bogus_Filename_Produces_an_Error")
        except TasmanianSG.TasmanianInputError as TSGError:
            TSGError.bShowOnExit = False
            self.assertEqual(TSGError.sVariable, "sFilename", "Reading a bogus file properly failed, but the error information is wrong.")

    def checkGlobalGridCustom(self):
        '''
        Test makeGlobalGridCustom(), which creates a grid from a CustomTabulated instance.
        '''
        gridA = TasmanianSG.makeGlobalGrid(1, 1, 3, "level", "custom-tabulated", [], 0.0, 0.0, tdata.sGaussPattersonTableFile)
        ct = TasmanianSG.makeCustomTabulatedFromFile(tdata.sGaussPattersonTableFile)
        gridB = TasmanianSG.makeGlobalGridCustom(1, 1, 3, "level", ct)
        ttc.compareGrids(gridA, gridB)
        gridA = TasmanianSG.makeGlobalGrid(2, 1, 3, "level", "custom-tabulated", [], 0.0, 0.0, tdata.sGaussPattersonTableFile)
        gridB = TasmanianSG.makeGlobalGridCustom(2, 1, 3, "level", ct)
        ttc.compareGrids(gridA, gridB)

    def performIOTest(self):
        self.checkMetaIO()

        self.checkReadWriteGlobal()
        self.checkReadWriteSequence()
        self.checkReadWriteLocalp()
        self.checkReadWriteWavelet()
        self.checkReadWriteFourier()

        self.checkCopySubgrid()
        self.checkReadWriteMisc()

        self.checkReadWriteCustomTabulated()
        self.checkGlobalGridCustom()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

# Using a single configuration file with all paths and relevant options
# The rest of the Tasmanian Python modules will not be configured by cmake

__version__ = "@Tasmanian_VERSION_MAJOR@.@Tasmanian_VERSION_MINOR@"
__license__ = "@Tasmanian_license@"
__author__ = "Miroslav Stoyanov"
__git_commit_hash__ = "@Tasmanian_git_hash@"

__path_libsparsegrid__ = "@Tasmanian_libsparsegrid_path@"
__path_libdream__      = "@Tasmanian_libdream_path@"
__path_libcaddons__    = "@Tasmanian_libcaddons_path@"
__path_logfile__       = "@Tasmanian_final_install_path@/share/Tasmanian/Tasmanian.log"

enableVerboseErrors = True

# the transitive library dependencies on Linux/OsX are handled by rpath, Windows needs help
if "@CMAKE_SYSTEM_NAME@" == "Windows":
    import os
    os.environ['PATH'] = "@Tasmanian_libpaths@" + os.pathsep + os.environ['PATH']

class TasmanianInputError(Exception):
    '''Exception raised for incorrect input to Tasmanian

    Attributes:
    sVariable   -- string containing the variable name with incorrect value
    sMessage    -- message regarding the error
    bShowOnExit -- boolean indicates whether to call printInfo() form the destructor

    '''
    def __init__(self, sVar, sMess):
        self.sVariable = sVar
        self.sMessage = sMess
        self.bShowOnExit = enableVerboseErrors

    def __del__(self):
        if self.bShowOnExit:
            self.printInfo()

    def printInfo(self):
        '''
        prints information for the incorect function or variable

        '''
        print("Incorrect input for: {0:s}".format(self.sVariable))
        print(self.sMessage)



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
import Tasmanian

def example_10():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 10: comparison between local polynomial and wavelet grids\n")

    iNumInputs = 2 # using two inputs for testing

    # test the error on a uniform dense grid with 10K points
    iTestGridSize = 100
    dx = np.linspace(-1.0, 1.0, iTestGridSize) # sample on a uniform grid
    aMeshX, aMeshY = np.meshgrid(dx, dx)
    aTestPoints = np.column_stack([aMeshX.reshape((iTestGridSize**2, 1)),
                                   aMeshY.reshape((iTestGridSize**2, 1))])

    def get_error(grid, model, aTestPoints):
        aGridResult = grid.evaluateBatch(aTestPoints)
        aModelResult = np.empty((aTestPoints.shape[0], 1), np.float64)
        for i in range(aTestPoints.shape[0]):
            aModelResult[i,:] = model(aTestPoints[i,:])
        return np.max(np.abs(aModelResult[:,0] - aGridResult[:,0]))

    def sharp_model(aX):
        return np.ones((1,)) * aX[0] / (1.0 + 100.0 * np.exp(-10.0 * aX[1]))

    grid_poly = Tasmanian.makeLocalPolynomialGrid(iNumInputs, 1, 3,
                                                     iOrder = 1, sRule = "localp")
    grid_wavelet = Tasmanian.makeWaveletGrid(iNumInputs, 1, 1, 1)

    print("            polynomial              wavelet")
    print("  points         error  points        error")

    fTolerance = 1.E-5;
    while((grid_poly.getNumNeeded() > 0) or (grid_wavelet.getNumNeeded() > 0)):
        Tasmanian.loadNeededValues(lambda x, tid: sharp_model(x), grid_poly, 4);
        Tasmanian.loadNeededValues(lambda x, tid: sharp_model(x), grid_wavelet, 4);

        # print the results at this stage
        print("{0:>8d}{1:>14.4e}{2:>8d}{3:>14.4e}".format(
            grid_poly.getNumLoaded(), get_error(grid_poly, sharp_model, aTestPoints),
            grid_wavelet.getNumLoaded(), get_error(grid_wavelet, sharp_model, aTestPoints)))

        # setting refinement for each grid
        grid_poly.setSurplusRefinement(fTolerance, 0, "fds");
        grid_wavelet.setSurplusRefinement(fTolerance, 0, "fds");


if (__name__ == "__main__"):
    example_10()
